/*


*/
#include <windows.h>
#include <stdio.h>

/* error function prototype */
void printError(char* functionName);

/* thread function prototypes */
DWORD WINAPI thread1(LPVOID);
DWORD WINAPI thread2(LPVOID);

HANDLE hPipe_1_In_Wr;  /* The first pipe's input and output handles. */
HANDLE hPipe_1_Out_Rd;
HANDLE hPipe_2_In_Wr;  /* The second pipe's input and output handles. */
HANDLE hPipe_2_Out_Rd;

int main( )
{
   // First Child.
   PROCESS_INFORMATION processInfo_1;

   STARTUPINFO startInfo_1;
   ZeroMemory(&startInfo_1, sizeof(STARTUPINFO));
   startInfo_1.cb = sizeof(STARTUPINFO);

   /* Create a pipe to connect to the first child's stdout. */
   if (! CreatePipe(&hPipe_1_Out_Rd, &hPipe_1_In_Wr, NULL, 0))
   {
      printError( "CreatePipe(1)" );
   }
   /* The write handle to the first pipe is needed by the
      first child, so it should be inherited by the child. */
   if (! SetHandleInformation(hPipe_1_In_Wr, HANDLE_FLAG_INHERIT, HANDLE_FLAG_INHERIT))
   {
      printError( "SetHandleInformation(hPipe_1_In_Wr)" );
   }
   /* The parent redirects the 1st child's stdout to the 1st pipe.
      This is done using the child's STARTUPINFO data structure.
      This redirection would not work if we had not set the new
      handle as inheritable. */
   startInfo_1.dwFlags |= STARTF_USESTDHANDLES;
   startInfo_1.hStdOutput = hPipe_1_In_Wr;
   startInfo_1.lpTitle = "child 1";
   /* Set up the child's command line. */
   LPTSTR lpCommandLine_1 = "echo.exe";
   /* Create the first child process. */
   if( !CreateProcess(NULL, lpCommandLine_1, NULL, NULL, TRUE,
                      CREATE_NEW_CONSOLE,
                      NULL, NULL, &startInfo_1, &processInfo_1) )
   {
      char buf[100];
      sprintf(buf, "CreateProcess(%s)", lpCommandLine_1);
      printError( buf );
   }
   else
   {
      fprintf(stderr, "Started process %d\n", (int)processInfo_1.dwProcessId);
   }
   /* Now that the first child has been created, we should close the
      pipe's input handle that is no longer needed by the parent. */
   if (! CloseHandle(hPipe_1_In_Wr))
   {
      printError("CloseHandle(hPipe_1_In_Wr)");
   }

   // Second Child.
   PROCESS_INFORMATION processInfo_2;

   STARTUPINFO startInfo_2;
   ZeroMemory(&startInfo_2, sizeof(STARTUPINFO));
   startInfo_2.cb = sizeof(STARTUPINFO);

   /* Create a pipe to connect to the second child's stdout. */
   if (! CreatePipe(&hPipe_2_Out_Rd, &hPipe_2_In_Wr, NULL, 0))
   {
      printError( "CreatePipe(2)" );
   }
   /* The write handle to the second pipe is needed by the
      second child, so it should be inherited by the child. */
   if (! SetHandleInformation(hPipe_2_In_Wr, HANDLE_FLAG_INHERIT, HANDLE_FLAG_INHERIT))
   {
      printError( "SetHandleInformation(hPipe_2_In_Wr)" );
   }
   /* The parent redirects the 2nd child's stdout to the 2nd pipe.
      This is done using the child's STARTUPINFO data structure.
      This redirection would not work if we had not set the new
      handle as inheritable. */
   startInfo_2.dwFlags |= STARTF_USESTDHANDLES;
   startInfo_2.hStdOutput = hPipe_2_In_Wr;
   startInfo_2.lpTitle = "child 2";
   /* Set up the child's command line. */
   LPTSTR lpCommandLine_2 = "echo.exe";
   /* Create the second child process. */
   if( !CreateProcess(NULL, lpCommandLine_2, NULL, NULL, TRUE,
                      CREATE_NEW_CONSOLE,
                      NULL, NULL, &startInfo_2, &processInfo_2) )
   {
      char buf[100];
      sprintf(buf, "CreateProcess(%s)", lpCommandLine_2);
      printError( buf );
   }
   else
   {
      fprintf(stderr, "Started process %d\n", (int)processInfo_2.dwProcessId);
   }
   /* Now that the second child has been created, we should close the
      pipe's input handle that is no longer needed by the parent. */
   if (! CloseHandle(hPipe_2_In_Wr))
   {
      printError("CloseHandle(hPipe_2_In_Wr)");
   }


   HANDLE hthread[2];

   // create the threads
   hthread[0] = CreateThread(NULL, 0, thread1, 0, 0, NULL);  // should have error checking
   hthread[1] = CreateThread(NULL, 0, thread2, 0, 0, NULL);

   char c;
   int n, counter = 0;
   HANDLE hStdIn = GetStdHandle(STD_INPUT_HANDLE);
   while (ReadFile(hStdIn, &c, 1, (DWORD*)&n, NULL) && n > 0)  // weird test for eof
   {
      counter += n;
      printf("%c", c);
   }
   printf("eof 3, counter = %d\n", counter);

   // wait for the threads to finish
   WaitForMultipleObjects(2, hthread, TRUE, INFINITE);


   /* Make the parent wait for the child process to end. */
   if ( WAIT_FAILED == WaitForSingleObject(processInfo_1.hProcess, INFINITE) )
   {
      printError("WaitForSingleObject(1)");
   }
   if ( WAIT_FAILED == WaitForSingleObject(processInfo_2.hProcess, INFINITE) )
   {
      printError("WaitForSingleObject(2)");
   }

   /* Close the handles we opened. */
   CloseHandle(hStdIn);
   CloseHandle(hthread[0]);
   CloseHandle(hthread[1]);
   CloseHandle(processInfo_1.hProcess);
   CloseHandle(processInfo_1.hThread);
   CloseHandle(processInfo_2.hProcess);
   CloseHandle(processInfo_2.hThread);

   return 0;
}


DWORD WINAPI thread1(LPVOID vargp) // vargp means "variable (length) argument pointer"
{
   char c;
   int n, counter = 0;
   while (ReadFile(hPipe_1_Out_Rd, &c, 1, (DWORD*)&n, NULL) && n > 0)  // weird test for eof
   {
      counter += n;
      printf("%c", c);
   }
   printf("eof 1, counter = %d\n", counter);
   return 0;
}


DWORD WINAPI thread2(LPVOID vargp)
{
   char c;
   int n, counter = 0;
   while (ReadFile(hPipe_2_Out_Rd, &c, 1, (DWORD*)&n, NULL) && n > 0)  // weird test for eof
   {
      counter += n;
      printf("%c", c);
   }
   printf("eof 2, counter = %d\n", counter);
   return 0;
}


/****************************************************************
   The following function can be used to print out "meaningful"
   error messages. If you call a Win32 function and it returns
   with an error condition, then call this function right away and
   pass it a string containing the name of the Win32 function that
   failed. This function will print out a reasonable text message
   explaining the error and then (if chosen) terminate the program.
*/
void printError(char* functionName)
{
   LPVOID lpMsgBuf;
    int error_no;
    error_no = GetLastError();
    FormatMessage(
         FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
         NULL,
         error_no,
         MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
         (LPTSTR) &lpMsgBuf,
         0,
         NULL
    );
    // Display the string.
    fprintf(stderr, "\n%s failed on error %d: ", functionName, error_no);
    fprintf(stderr, (const char*)lpMsgBuf);
    // Free the buffer.
    LocalFree( lpMsgBuf );
    ExitProcess(1);  // terminate the program
}//printError
